
import os
import typing
from ._base import Filelike

class JoinedFiles(Filelike):
    class _RelativeOffset:
        def __init__(self, idx:int, offset:int):
            self.idx = idx
            self.offset = offset

    class PartInfo:
        def __init__(self, idx:int, offset:int, size:int):
            self.idx = idx
            self.offset = offset
            self.size = size

    def __init__(self, parts:'typing.List[Filelike]'):
        super().__init__(moduleName='join')
        self.parts = parts
        self._pos:int = 0
        self._size:int = sum(part.size() for part in parts)
        self._offsets:typing.List[int] = []
        self._currentIndex = 0
        offset = 0
        for part in parts:
            self._offsets.append(offset)
            offset += part.size()

    def _getRelativeOffset(self, offset:int) -> _RelativeOffset:
        # check current part first
        idx = self._currentIndex
        partOffset = self._offsets[idx]
        partSize = self.parts[idx].size()
        if offset >= partOffset and offset < partOffset+partSize:
            return JoinedFiles._RelativeOffset(idx=idx, offset=offset-partOffset)

        # no match -> check other parts
        for i, partOffset in enumerate(self._offsets):
            if partOffset > offset: break
            idx = i

        partOffset = self._offsets[idx]
        partSize = self.parts[idx].size()
        return JoinedFiles._RelativeOffset(idx=idx, offset=offset-partOffset)

    def __repr__(self) -> str:
        return f"JoinedFiles({', '.join([repr(part) for part in self.parts])})"

    def getPositionInfo(self, pos: int) -> typing.Tuple[str, int]:
        info = self._getRelativeOffset(pos)
        part = self.parts[info.idx]
        return part.getPositionInfo(info.offset)

    def read(self, n: int = -1) -> bytes:
        if n < 0: n = self._size - self._pos
        if n <= 0: return b''

        rc = b''
        info = self._getRelativeOffset(self._pos)
        while True:
            f = self.parts[info.idx]
            f.seek(info.offset)
            data = f.read(n - len(rc))
            if not data: break
            rc += data
            self._pos += len(data)
            info = self._getRelativeOffset(self._pos)
            self._currentIndex = info.idx
        return rc

    def seek(self, offset: int, whence: int = os.SEEK_SET) -> int:
        if whence == os.SEEK_SET:
            self._pos = offset
        elif whence == os.SEEK_CUR:
            self._pos += offset
        elif whence == os.SEEK_END:
            self._pos = self._size + offset
        else: raise ValueError(f"seek(): unsupported whence param: {whence}")

        if self._pos < 0: self._pos = 0
        if self._pos > self._size: self._pos = self._size

        # find the part that matches our position
        idx = 0
        for i, offset in enumerate(self._offsets):
            if self._pos < offset: break
            idx = i
        self._currentIndex = idx

        return self._pos
 
    def size(self) -> int: return self._size
    def tell(self) -> int: return self._pos

def join(*parts: Filelike):
    " combines filelike objects into a seamless whole, i.e. read() calls should not stop at part boundaries "
    return JoinedFiles(list(parts))



import os
import snipfile


def test_join():
    hello = snipfile.fromBytes(b"hello\n")
    world = snipfile.fromBytes(b"world\n")

    helloworld = snipfile.join(hello, world)
    assert helloworld.read() == b'hello\nworld\n'
    assert snipfile.join(world, hello, hello, world).read() == b'world\nhello\nhello\nworld\n'

    empty = snipfile.Slice(helloworld, offset=4, size=0)
    assert snipfile.join(hello, empty, world).read() == b'hello\nworld\n'

    assert snipfile.join(*snipfile.splitAfter(helloworld, b'\n')).read() == b'hello\nworld\n'

    assert snipfile.join(empty, empty, world).read() == b'world\n'

def test_join_empty():
    nothingness = snipfile.join()
    assert nothingness.size() == 0
    assert nothingness.read() == b''

    # JoinedFiles will correct seek positions
    assert nothingness.seek(100) == 0
    assert nothingness.seek(1, whence=os.SEEK_CUR) == 0
    assert nothingness.seek(-100, os.SEEK_CUR) == 0
    assert nothingness.seek(100, os.SEEK_END) == 0
    assert nothingness.seek(-100, os.SEEK_END) == 0

    assert nothingness.read() == b''
    assert nothingness.tell() == 0


import os
import typing

CHUNK_SIZE=8192

from ._base import Filelike

class Slice(Filelike):
    """ represents a smaller slice of another fileobj, only giving access to the given section """
    def __init__(self, f: Filelike, *, offset:int=0, size:typing.Optional[int]=None):
        if size is None: size = f.size()
        if offset < 0: offset = size + offset # subtract offset from size (offset is negative)
        if offset < 0: offset = 0

        if offset > f.size(): offset = f.size()
        if offset + size > f.size(): size = f.size() - offset

        self.f = f
        self.offset = offset # start of our 'window' into the file (which is of size `size`)
        self._pos = 0 # position we're reading right now
        self._size = size

    def __repr__(self) -> str:
        return f"Slice(offset={self.offset}, size={self.size()}, f={repr(self.f)})"

    def getPositionInfo(self, pos: int) -> typing.Tuple[str, int]:
        return self.f.getPositionInfo(self.offset + pos)

    def read(self, n:int=-1) -> bytes:
        realpos = self.f.seek(self.offset + self._pos, os.SEEK_SET)
        pos = realpos - self.offset
        if n < 0: n = self._size-pos
        if pos+n > self._size: n = self._size-pos
        rc = self.f.read(n)
        self._pos = pos + len(rc)
        return rc
    
    def seek(self, offset:int, whence:int=os.SEEK_SET) -> int:
        if whence == os.SEEK_SET:
            self._pos = offset
        elif whence == os.SEEK_CUR:
            self._pos += offset
        elif whence == os.SEEK_END:
            self._pos = self.size() + offset
        else: raise ValueError(f'seek(): invalid whence: {repr(whence)}')

        if self._pos < 0: self._pos = 0
        if self._pos > self.size(): self._pos = self.size()
        return self._pos

    def size(self) -> int: return self._size
    def tell(self) -> int: return self._pos

def _split(f: Filelike, delimiter: bytes, *, bytesAfter:int=0, emptyTail:bool=True) -> typing.Generator[Slice,None,None]:
    if not delimiter: raise ValueError("split(): delimiter has to be nonempty")
    if len(delimiter) > CHUNK_SIZE/2: raise ValueError('delimiter too long')
    data = b''
    dataOffset = 0
    oldOffset = -1 # if the offset doesn't change between iterations, we'll abort
    sliceStart = 0
    while True:
        if dataOffset == oldOffset: break
        f.seek(dataOffset, os.SEEK_SET)
        oldOffset = dataOffset

        data = f.read(CHUNK_SIZE)
        if not data: break
        try:
            idx = data.index(delimiter)
        except ValueError:
            if len(data)-len(delimiter) == 0:
                break # we're at the end
            dataOffset += max(0, len(data)-len(delimiter))
            continue # not found

        sliceLen = idx
        yield Slice(f, offset=sliceStart, size=sliceLen+bytesAfter)
        dataOffset += idx+len(delimiter)
        sliceStart = dataOffset
    if emptyTail or f.size() != sliceStart:
        # there's data left after the last delimiter
        yield Slice(f, offset=sliceStart)

def cutAt(f:Filelike, *positions:int) -> typing.List[Slice]:
    """ cuts a file into Slices at the given positions, e.g.:
cutAt(fromBytes(b"hello you", 4,6)) returns a list of Slices, one for b'hell', one for b'o ' and one for b'you'

use it e.g. as:

```
twoParts = cutAt(f, 32) # cut at position 32, returning a list of Slices
hell, o_, you = cutAt(fromBytes(b"hello_you"), 4,6)
```

- will always try to return len(positions)+1 slices
- raises ValueError if the cut positions aren't in ascending order """
    rc:typing.List[Slice] = []
    lastCut = 0
    for cut in positions:
        if cut < 0: # relative to the end of the file
            cut = f.size() + cut
        if cut < lastCut: raise ValueError(f"cuts are not in order")
        rc.append(Slice(f, offset=lastCut, size=cut-lastCut))
        lastCut = cut
    rc.append(Slice(f, offset=lastCut))
    return rc


def split(f:Filelike, delimiter:bytes):
    return _split(f, delimiter)
    
def splitAfter(f:Filelike, delimiter:bytes):
    return _split(f, delimiter, bytesAfter=len(delimiter), emptyTail=False)



import io
import os
import typing

from ._base import Filelike

class File(Filelike):
    " thin wrapper around python file-like objects "
    def __init__(self, fileobj:typing.Union[typing.BinaryIO, str]):
        super().__init__(moduleName='file')
        if isinstance(fileobj, str):
            fileobj = open(fileobj, 'rb')
        self._size = fileobj.seek(0, os.SEEK_END)
        fileobj.seek(0)
        self.f = fileobj
        #self.name = fileobj.name

    def getPositionInfo(self, pos: int) -> typing.Tuple[str,int]:
        return self.f.name, pos

    def read(self, n:int=-1) -> bytes: return self.f.read(n)
    def seek(self, offset: int, whence:int=os.SEEK_SET) -> int:
        return self.f.seek(offset, whence)
    def size(self) -> int: return self._size
    def tell(self) -> int: return self.f.tell()

def fromBytes(data:bytes) -> File:
    " syntactic sugar for `File(io.BytesIO(data))` "
    return File(io.BytesIO(data))


import os

import snipfile

def test_slice():
    # default values (i.e. the full file)
    slice = snipfile.Slice(snipfile.fromBytes(b'hello world'))
    assert slice.read() == b'hello world'
    assert slice.offset == 0
    assert slice.size() == 11

    # offset only
    slice = snipfile.Slice(snipfile.fromBytes(b'hello world'), offset=3)
    assert slice.read() == b'lo world'
    assert slice.offset == 3
    assert slice.size() == 8

    # size only
    slice = snipfile.Slice(snipfile.fromBytes(b'hello world'), size=4)
    assert slice.read() == b'hell'
    assert slice.offset == 0
    assert slice.size() == 4

    # both
    slice = snipfile.Slice(snipfile.fromBytes(b'hello world'), offset=3, size=5)
    assert slice.read() == b'lo wo'
    assert slice.offset == 3
    assert slice.size() == 5

    # negative offset (i.e. last 5 bytes of the file)
    slice = snipfile.Slice(snipfile.fromBytes(b'hello world'), offset=-5)
    assert slice.read() == b'world'
    assert slice.offset == 6
    assert slice.size() == 5

def test_seek():
    slice = snipfile.Slice(snipfile.fromBytes(b'hello, my dear world'), offset=7, size=7) # "my dear"
    assert slice.read() == b'my dear'
    assert slice.read() == b''

    slice.seek(0, os.SEEK_SET)
    assert slice.read(2) == b'my'
    assert slice.read(1) == b' '
    assert slice.read(6) == b'dear'

    slice.seek(-99, os.SEEK_SET)
    assert slice.tell() == 0
    slice.seek(99, os.SEEK_SET)
    assert slice.tell() == 7

    slice.seek(-4, os.SEEK_CUR)
    assert slice.read() == b'dear'
    slice.seek(-99, os.SEEK_CUR)
    assert slice.tell() == 0

    slice.seek(0, os.SEEK_END)
    assert slice.tell() == 7
    slice.seek(4, os.SEEK_END)
    assert slice.tell() == 7
    slice.seek(-4, os.SEEK_END)
    assert slice.read() == b'dear'



def test_split():
    data = b'a;b'
    parts = [slice.read() for slice in snipfile.split(snipfile.fromBytes(data), b';')]
    assert parts == data.split(b';')

    data = b';'
    parts = [slice.read() for slice in snipfile.split(snipfile.fromBytes(data), b';')]
    assert parts == data.split(b';')

    data = b';;'
    parts = [slice.read() for slice in snipfile.split(snipfile.fromBytes(data), b';')]
    assert parts == data.split(b';')

    data = b'hello world, this.is.a test '
    parts = [slice.read() for slice in snipfile.split(snipfile.fromBytes(data), b" ")]
    assert parts == data.split(b' ')

    data = b'bananabread'
    parts = [slice.read() for slice in snipfile.split(snipfile.fromBytes(data), b'an')]
    assert parts == data.split(b'an')

def test_split_long_delimiter():
    # delimiter's longer than the file
    data = b'abc'
    parts = [slice.read() for slice in snipfile.split(snipfile.fromBytes(data), b'thisislongerthanbuff')]
    assert parts == [b'abc']

    data = b''
    parts = [slice.read() for slice in snipfile.split(snipfile.fromBytes(data), b';')]
    assert parts == [b'']

def test_splitAfter():
    data = b'hello\nworld\n'
    parts = [slice.read() for slice in snipfile.splitAfter(snipfile.fromBytes(data), b'\n')]
    assert parts == [b'hello\n', b'world\n']

def test_cut():
    data = b'hello world'
    parts = [slice.read() for slice in snipfile.cutAt(snipfile.fromBytes(data), 2,5,6)]
    assert parts == [b'he', b'llo', b' ', b'world']

    parts = [slice.read() for slice in snipfile.cutAt(snipfile.fromBytes(data), 2,-2)]
    assert parts == [b'he', b'llo wor', b'ld']

    # we expect to get empty slices in stead of (omitting empty sections)
    parts = [slice.read() for slice in snipfile.cutAt(snipfile.fromBytes(data), 6, 10)]
    assert parts == [b'hello ', b'worl', b'd']
    parts = [slice.read() for slice in snipfile.cutAt(snipfile.fromBytes(data), 6, 11)]
    assert parts == [b'hello ', b'world', b'']
    parts = [slice.read() for slice in snipfile.cutAt(snipfile.fromBytes(data), 6, 6, 11)]
    assert parts == [b'hello ', b'', b'world', b'']

    # providing cuts in the wrong order leads to an Exception
    try:
        parts = [slice.read() for slice in snipfile.cutAt(snipfile.fromBytes(data), 8,6,4,2)]
        assert False, "we expect the code to raise a ValueError if the user provides cut points in the wrong order"
    except ValueError:
        pass




import abc
import os
import typing

if hasattr(typing, 'Protocol'):
    class FileIntf(typing.Protocol):
        def getPositionInfo(self, pos:int): ...

        def read(self, n:int=-1) -> bytes: ...
        def seek(self, offset:int, whence:int=os.SEEK_SET) -> int: ...
        def size(self) -> int: ...
        def tell(self) -> int: ...

class Filelike(abc.ABC):
    " base class for all our file classes"
    def __init__(self, *, moduleName:str) -> None:
        self.moduleName = moduleName

    @abc.abstractmethod
    def getPositionInfo(self, pos:int) -> typing.Tuple[str,int]:
        """ translates pos into the path and absolute position in the underlying file; can be used to provide valuable position info to the user """

    @abc.abstractmethod
    def read(self, n:int=-1) -> bytes: ...

    @abc.abstractmethod
    def seek(self, offset:int, whence:int=os.SEEK_SET) -> int:
        """ seek() to the given position, relative to `whence`, e.g.:
- seek(n) behaves like seek(0, os.SEEK_SET)
- seek(n, os.SEEK_SET) moves the file cursor to position 100"
- seek(n, os.SEEK_CUR) moves the file cursor to tell()+n
- seek(n, os.SEEK_END) moves the file cursor to size()+n
"""

    @abc.abstractmethod
    def size(self) -> int: ...

    @abc.abstractmethod
    def tell(self) -> int: ...

    def writeTo(self, target:typing.BinaryIO):
        " writes the whole contents of this file or slice to target "
        self.seek(0)
        while True:
            data = self.read(8192)
            if not data: break
            target.write(data)



import typing
from . import _base
from . import _join
from . import _slice

@typing.overload
def punchHole(f: _base.Filelike, *, start:int, size:int) -> _base.Filelike: ...
@typing.overload
def punchHole(f: _base.Filelike, *, start:int, end:int) -> _base.Filelike: ...


def punchHole(f: _base.Filelike, *, start:int, size:typing.Optional[int]=None, end:typing.Optional[int]=None) -> _base.Filelike:
    """ take out a section of f (by combining cutAt() and join())
you can specify the hole either by its start and size or by a start+end position.

- will raise ValueError if start<0, size<0 or end<start
- if you provide both or neither of end and size, we'll also raise a ValueError
- if you attempt to punch an empty hole (i.e. size=0 or start==end), f will be returned unmodified"""
    if end is None:
        if size is None: 
            raise ValueError("please provide a size or end to snipfile.punchHole()")
        end = start+size
    elif size is not None:
        raise ValueError("snipfile.punchHole() only accepts one of size/end")

    size = end-start
    if size < 0: raise ValueError("punchHole(): got hole of negative hole size")
    if start < 0: raise ValueError(f"punchHole() got negative start={repr(start)}")
    if size == 0: return f #shortcut

    before, hole_, after = _slice.cutAt(f, start, end)
    assert hole_.size() == size
    return _join.join(before, after)